==> healthrian_ble_interface-1.29.0.tar.gz: lib/src/module/module_oxi.dart <==
import 'package:copy_with_extension/copy_with_extension.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extension/extension.dart';
import 'package:lzma/lzma.dart';

part 'module_oxi.g.dart';

@CopyWith()
class ModuleOxi {
  ModuleOxi({
    required this.ble,
    this.commands = const Commands(),
    this.udfConvert = Builtin.udfConvertWearCardioT,
    this.samplingRate = 100,
  });

  final ModuleBle ble;
  final Commands commands;

  final List<List<double>>? Function({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) udfConvert;

  /////

  final int samplingRate;

  int sampleCount = 30;
  int channelCount = 2;

  /// [maxBuffer] decides the maximum length of buffered Stream.
  double maxBuffer = 40;

  int get _maxBuffer => (maxBuffer * samplingRate).round();

  /// [lengthBuffer] decides a length of buffered Stream,
  double lengthBuffer = 30;

  int get _lengthBuffer => (lengthBuffer * samplingRate).round();

  /// [displayBuffer] decides a length of buffered Stream,
  /// which is shown to screen.
  double displayBuffer = 5;

  int get _displayBuffer => (displayBuffer * samplingRate).round();

  /// [refreshBuffer] decides a length to be hidden after refreshed.
  double refreshBuffer = 0.5;

  int get _refreshBuffer => (refreshBuffer * samplingRate).round();

  ///////////////////////////////////////////////////

  /// [isFiltered] decides whether ecg signal is filtered.
  bool isFiltered = true;

  /// If [isCompressed] is `true`,
  /// pulse oxi signal is decompressed by LZMA algorithm.
  bool isCompressed = false;

  DataFilter dataFilter = DataFilter(1);

  AdaptiveHeartRate adaptiveHR = AdaptiveHeartRate();

  /// [callbacks] is functions to be executed periodically.
  final List<(void Function(Channels, DataFilter)? callback, int period)>
      callbacks = [];

  /// [lastChannel] is a latest stream value from [ble.streams.bufferedOxi]
  Channels? lastChannel;

  bool _isPaused = false;
  bool _isMeasuring = false;

  void initStream() {
    ble.streams
      ..rawOxi = null
      ..lzmaOxi = null
      ..convertedOxi = null
      ..eventOxi = null
      ..bufferedOxi = null
      ..bufferedFilterOxi = null
      ..plotOxi = null
      ..plotRTOxi = null;
  }

  /// [start] sends start_oxi_packet to device.
  Future<void> start() async {
    await _getConfig();
    await _sendStart();

    _isMeasuring = true;
    _isPaused = false;

    ble.streams
      ..rawOxi = _getRaw().broadcast
      ..lzmaOxi = _getLzma().broadcast
      ..eventOxi = _getEvent()?.broadcast
      ..convertedOxi = _getConverted().broadcast
      ..bufferedOxi = _getBuffered().broadcast
      ..bufferedFilterOxi = _getBufferedFilter().broadcast
      ..plotOxi = _getPlot().broadcast
      ..plotRTOxi = _getPlotRT().broadcast;
  }

  /// [pause]
  Future<void> pause() async {
    _isPaused = !_isPaused;
  }

  /// [stop] sends stop_oxi_packet to device.
  Future<void> stop() async {
    _isMeasuring = false;
    final control = ble.points[Characteristics.STOP_OXI];
    await control.writePacket(commands.stopOxi);
  }

  /////////////////////////

  Future<bool> _checkValid() async {
    final confirm = ble.points[Characteristics.CONTROL_POINT_RESPONSE];
    if (confirm == null) return false;

    final confirmValue = await confirm.read();
    if (confirmValue.isEmpty) return false;
    if (confirmValue[0] == 2) return false;
    return true;
  }

  Future<void> _getConfig() async {
    final configuration = ble.points[Characteristics.CONFIG_OXI];
    if (configuration == null) return;

    final configurationValue = await configuration.read();

    if (configurationValue.length < 2) return;

    channelCount = configurationValue[0];
    sampleCount = configurationValue[1];
  }

  Future<void> _sendStart() async {
    final control = ble.points[Characteristics.START_OXI];
    await control.writePacket(commands.startEcg);
  }

  ////////////////////////////

  Stream<List<int>> _getRaw() async* {
    final measurement = ble.points[Characteristics.MEASUREMENT_OXI];
    if (measurement == null) return;
    try {
      await measurement.setNotifyValue(true);
    } catch (e) {
      '_getStream : $e'.log;
    }

    await for (final res in measurement.lastValueStream) {
      if (!_isMeasuring) break;
      if (!ble.isSimulator && ble.device == null) break;
      if (ble.isSimulator && !ble.isSimulatorConnected) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getLzma() async* {
    final events = ble.streams.rawOxi;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? _getEvent() async* {
    final events = isCompressed ? ble.streams.lzmaOxi : ble.streams.rawOxi;
    if (events == null) return;
    await for (final event in events) {
      yield event;
    }
  }

  Stream<List<List<double>>> _getConverted() async* {
    final events = ble.streams.eventOxi;
    if (events == null) return;
    await for (final event in events) {
      final value = udfConvert(
        event: event,
        channelCount: channelCount,
        sampleCount: sampleCount,
      );
      if (value == null) continue;
      yield value;
    }
  }

  Stream<Channels> _getBuffered() async* {
    final events = ble.streams.convertedOxi;
    if (events == null) return;

    final res = List.generate(channelCount, (index) => <double>[]);

    var x = 0.0;
    final callbacksTargets = [for (final callback in callbacks) callback.$2];

    await for (final event in events) {
      for (var k = 0; k < channelCount; k++) {
        res[k].addAll(event[k]);
      }

      for (final channel in res) {
        if (channel.length > _maxBuffer) {
          channel.removeRange(0, channel.length - _lengthBuffer);
        }
      }

      final channels = Channels.fromList(res);
      yield channels;

      for (var i = 0; i < callbacks.length; i++) {
        final callback = callbacks[i].$1;
        final period = callbacks[i].$2;
        if (callback != null && x >= callbacksTargets[i]) {
          callback.call(channels, dataFilter);
          callbacksTargets[i] += period;
        }
      }

      lastChannel = channels;

      x += sampleCount;
    }
  }

  // TODO: implement filtering for pulse oxi
  Stream<Channels> _getBufferedFilter() async* {
    final events = ble.streams.bufferedEcg;
    if (events == null) return;

    await for (final event in events) {
      yield event;
    }
  }

  Stream<PlotChannels> _getPlot() async* {
    final events =
        isFiltered ? ble.streams.bufferedFilterEcg : ble.streams.bufferedEcg;

    if (events == null) return;

    final resList = List.generate(channelCount, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0, _refreshBuffer, sampleCount);
    var x = 0.0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        for (var k = 0; k < event.list.length; k++) {
          res.list[k].add(FlSpot(x + j, event.list[k][i]));
        }
      }

      if (res.l2.length > event.l2.length) {
        for (var k = 0; k < event.list.length; k++) {
          res.list[k].removeRange(0, res.list[k].length - event.list[k].length);
        }
      }

      yield res;
      x += sampleCount;
    }
  }

  Stream<PlotChannels?> _getPlotRT() async* {
    final events = ble.streams.plotEcg;

    if (events == null) return;

    double skip(double x) {
      final newX = x % _displayBuffer;
      final newXSkip = newX - newX;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    final resList = List.generate(
      channelCount,
      (_) => List<FlSpot>.generate(
        _displayBuffer,
        (index) => FlSpot.nullSpot,
        growable: false,
      ),
    );

    final res = PlotChannels.fromList(resList, 0, _refreshBuffer, sampleCount);
    var x = 0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        final index = skip(x + j.toDouble()).round();

        for (var k = 0; k < event.list.length; k++) {
          res.list[k][index] = f(event.list[k][i]);
        }

        if (_refreshBuffer > 0) {
          final refreshedIndex =
              skip(x + j.toDouble() + _refreshBuffer).round();

          for (var k = 0; k < event.list.length; k++) {
            res.list[k][refreshedIndex] = FlSpot.nullSpot;
          }
        }
      }

      x += sampleCount;

      yield res.copyWith(
        pivot: (x % _displayBuffer),
      );
    }
    yield null;
  }
}

==> healthrian_ble_interface-1.30.0.tar.gz: lib/src/module/module_oxi.dart <==
import 'package:copy_with_extension/copy_with_extension.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extension/extension.dart';
import 'package:lzma/lzma.dart';

part 'module_oxi.g.dart';

@CopyWith()
class ModuleOxi {
  ModuleOxi({
    required this.ble,
    this.commands = const Commands(),
    this.udfConvert = Builtin.udfConvertWearCardioT,
    this.samplingRate = 100,
  });

  final ModuleBle ble;
  final Commands commands;

  final List<List<double>>? Function({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) udfConvert;

  /////

  final int samplingRate;

  int sampleCount = 30;
  int channelCount = 2;

  /// [maxBuffer] decides the maximum length of buffered Stream.
  double maxBuffer = 40;

  int get _maxBuffer => (maxBuffer * samplingRate).round();

  /// [lengthBuffer] decides a length of buffered Stream,
  double lengthBuffer = 30;

  int get _lengthBuffer => (lengthBuffer * samplingRate).round();

  /// [displayBuffer] decides a length of buffered Stream,
  /// which is shown to screen.
  double displayBuffer = 5;

  int get _displayBuffer => (displayBuffer * samplingRate).round();

  /// [refreshBuffer] decides a length to be hidden after refreshed.
  double refreshBuffer = 0.5;

  int get _refreshBuffer => (refreshBuffer * samplingRate).round();

  ///////////////////////////////////////////////////

  /// [isFiltered] decides whether ecg signal is filtered.
  bool isFiltered = true;

  /// If [isCompressed] is `true`,
  /// pulse oxi signal is decompressed by LZMA algorithm.
  bool isCompressed = false;

  DataFilter dataFilter = DataFilter(1);

  AdaptiveHeartRate adaptiveHR = AdaptiveHeartRate();

  /// [callbacks] is functions to be executed periodically.
  final List<(void Function(Channels, DataFilter)? callback, int period)>
      callbacks = [];

  /// [lastChannel] is a latest stream value from [ble.streams.bufferedOxi]
  Channels? lastChannel;

  bool _isPaused = false;
  bool _isMeasuring = false;

  void initStream() {
    ble.streams.oxi.preload
      ..raw = null
      ..lzma = null
      ..event = null
      ..value = null;
    ble.streams.oxi.original
      ..buffer = null
      ..plot = null
      ..plotRT = null;
    ble.streams.oxi.filter
      ..buffer = null
      ..plot = null
      ..plotRT = null;
  }

  /// [start] sends start_oxi_packet to device.
  Future<void> start() async {
    await _getConfig();
    await _sendStart();

    _isMeasuring = true;
    _isPaused = false;

    ble.streams.oxi.preload
      ..raw = _getRaw().broadcast
      ..lzma = _getLzma().broadcast
      ..event = _getEvent()?.broadcast
      ..value = _getConverted().broadcast;
    ble.streams.oxi.original
      ..buffer = _getBuffered().broadcast
      ..plot = _getPlot().broadcast
      ..plotRT = _getPlotRT().broadcast;
    ble.streams.oxi.filter
      ..buffer = _getBufferedFilter().broadcast
      ..plot = null
      ..plotRT = null;
  }

  /// [pause]
  Future<void> pause() async {
    _isPaused = !_isPaused;
  }

  /// [stop] sends stop_oxi_packet to device.
  Future<void> stop() async {
    _isMeasuring = false;
    final control = ble.points[Characteristics.STOP_OXI];
    await control.writePacket(commands.stopOxi);
  }

  /////////////////////////

  // Future<bool> _checkValid() async {
  //   final confirm = ble.points[Characteristics.CONTROL_POINT_RESPONSE];
  //   if (confirm == null) return false;
  //
  //   final confirmValue = await confirm.read();
  //   if (confirmValue.isEmpty) return false;
  //   if (confirmValue[0] == 2) return false;
  //   return true;
  // }

  Future<void> _getConfig() async {
    final configuration = ble.points[Characteristics.CONFIG_OXI];
    if (configuration == null) return;

    final configurationValue = await configuration.read();

    if (configurationValue.length < 2) return;

    channelCount = configurationValue[0];
    sampleCount = configurationValue[1];
  }

  Future<void> _sendStart() async {
    final control = ble.points[Characteristics.START_OXI];
    await control.writePacket(commands.startEcg);
  }

  ////////////////////////////

  Stream<List<int>> _getRaw() async* {
    final measurement = ble.points[Characteristics.MEASUREMENT_OXI];
    if (measurement == null) return;
    try {
      await measurement.setNotifyValue(true);
    } catch (e) {
      '_getStream : $e'.log;
    }

    await for (final res in measurement.lastValueStream) {
      if (!_isMeasuring) break;
      if (!ble.isSimulator && ble.device == null) break;
      if (ble.isSimulator && !ble.isSimulatorConnected) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getLzma() async* {
    final events = ble.streams.oxi.preload.raw;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? _getEvent() async* {
    final events = isCompressed
        ? ble.streams.oxi.preload.lzma
        : ble.streams.oxi.preload.raw;
    if (events == null) return;
    await for (final event in events) {
      yield event;
    }
  }

  Stream<List<List<double>>> _getConverted() async* {
    final events = ble.streams.oxi.preload.event;
    if (events == null) return;
    await for (final event in events) {
      final value = udfConvert(
        event: event,
        channelCount: channelCount,
        sampleCount: sampleCount,
      );
      if (value == null) continue;
      yield value;
    }
  }

  Stream<Channels> _getBuffered() async* {
    final events = ble.streams.oxi.preload.value;
    if (events == null) return;

    final res = List.generate(channelCount, (index) => <double>[]);

    var x = 0.0;
    final callbacksTargets = [for (final callback in callbacks) callback.$2];

    await for (final event in events) {
      for (var k = 0; k < channelCount; k++) {
        res[k].addAll(event[k]);
      }

      for (final channel in res) {
        if (channel.length > _maxBuffer) {
          channel.removeRange(0, channel.length - _lengthBuffer);
        }
      }

      final channels = Channels.fromList(res);
      yield channels;

      for (var i = 0; i < callbacks.length; i++) {
        final callback = callbacks[i].$1;
        final period = callbacks[i].$2;
        if (callback != null && x >= callbacksTargets[i]) {
          callback.call(channels, dataFilter);
          callbacksTargets[i] += period;
        }
      }

      lastChannel = channels;

      x += sampleCount;
    }
  }

  // TODO: implement filtering for pulse oxi
  Stream<Channels> _getBufferedFilter() async* {
    final events = ble.streams.oxi.original.buffer;
    if (events == null) return;

    await for (final event in events) {
      yield event;
    }
  }

  Stream<PlotChannels> _getPlot() async* {
    final events = ble.streams.oxi.original.buffer;

    if (events == null) return;

    final resList = List.generate(channelCount, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0, _refreshBuffer, sampleCount);
    var x = 0.0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        for (var k = 0; k < event.list.length; k++) {
          res.list[k].add(FlSpot(x + j, event.list[k][i]));
        }
      }

      if (res.l2.length > event.l2.length) {
        for (var k = 0; k < event.list.length; k++) {
          res.list[k].removeRange(0, res.list[k].length - event.list[k].length);
        }
      }

      yield res;
      x += sampleCount;
    }
  }

  Stream<PlotChannels?> _getPlotRT() async* {
    final events = ble.streams.oxi.original.plot;

    if (events == null) return;

    double skip(double x) {
      final newX = x % _displayBuffer;
      final newXSkip = newX - newX;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    final resList = List.generate(
      channelCount,
      (_) => List<FlSpot>.generate(
        _displayBuffer,
        (index) => FlSpot.nullSpot,
        growable: false,
      ),
    );

    final res = PlotChannels.fromList(resList, 0, _refreshBuffer, sampleCount);
    var x = 0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        final index = skip(x + j.toDouble()).round();

        for (var k = 0; k < event.list.length; k++) {
          res.list[k][index] = f(event.list[k][i]);
        }

        if (_refreshBuffer > 0) {
          final refreshedIndex =
              skip(x + j.toDouble() + _refreshBuffer).round();

          for (var k = 0; k < event.list.length; k++) {
            res.list[k][refreshedIndex] = FlSpot.nullSpot;
          }
        }
      }

      x += sampleCount;

      yield res.copyWith(
        pivot: (x % _displayBuffer),
      );
    }
    yield null;
  }
}
